#ifndef CORE_CM4_H
#define CORE_CM4_H

#include <stdint.h>
#include <core_instructions.h>
//#include <core_functions.h>

#define __I volatile const
#define __O volatile
#define __IO volatile

/** Register Abstraction **/

/* Structure type to access the Nested Vector Interrupt Controller (NVIC). */
typedef struct {
	__IO uint32_t ISER[8]; // Offset: 0x000 (R/W) - Interrupt Set Enable Register
	uint32_t RESERVED0[24];
	__IO uint32_t ICER[8]; // Offset: 0x080 (R/W) - Interrupt Clear Enable Register
	uint32_t RESERVED1[24];
	__IO uint32_t ISPR[8]; // Offset: 0x100 (R/W) - Interrupt Set Pending Register
	uint32_t RESERVED2[24];
	__IO uint32_t ICPR[8]; // Offset: 0x180 (R/W) - Interrupt Clear Pending Register
	uint32_t RESERVED3[24];
	__IO uint32_t IABR[8]; // Offset: 0x200 (R/W) - Interrupt Active Bit Register
	uint32_t RESERVED4[56];
	__IO uint8_t IP[240]; // Offset: 0x300 (R/W) - Interrupt Priority Register (8-bit Wide)
	uint32_t RESERVED5[644];
	__O uint32_t STIR; // Offset: 0xE00 (W) - Software Trigger Interrupt Register
} NVIC_t;

/* Structure type to access the System Control Block (SCB). */
typedef struct {
	__I uint32_t CPUID; // Offset: 0x000 (R) - CPUID Base Register
	__IO uint32_t ICSR; // Offset: 0x004 (R/W) - Interrupt Control and State Register
	__IO uint32_t VTOR; // Offset: 0x008 (R/W) - Vector Table Offset Register
	__IO uint32_t AIRCR; // Offset: 0x00C (R/W) - Application Interrupt and Reset Controller
	__IO uint32_t SCR; // Offset: 0x010 (R/W) - System Control Register
	__IO uint32_t CCR; // Offset: 0x014 (R/W) - Configuration Control Register
	__IO uint8_t SHPR[12]; // Offset: 0x018 (R/W) - System Handlers Priority Registers (4-7, 8-11, 12-15)
	__IO uint32_t SHCSR; // Offset: 0x024 (R/W) - System Handler Control and State Register
	__IO uint32_t CFSR; // Offset: 0x028 (R/W) - Configurable Fault Status Register
	__IO uint32_t HFSR; // Offset: 0x02C (R/W) - HardFault Status Register
	__IO uint32_t DFSR; // Offset: 0x030 (R/W) - Debug Fault Status Register
	__IO uint32_t MMFAR; // Offset: 0x034 (R/W) - MemManage Fault Address Register
	__IO uint32_t BFAR; // Offset: 0x038 (R/W) - BusFault Address Register
	__IO uint32_t AFSR; // Offset: 0x03C (R/W) - Auxiliary Fault Status Register
	__I uint32_t PFR[2]; // Offset: 0x040 (R) - Processor Feature Register
	__I uint32_t DFR; // Offset: 0x048 (R) - Debug Feature Register
	__I uint32_t AFR; // Offset: 0x04C (R) - Auxiliary Feature Register
	__I uint32_t MMFR[4]; // Offset: 0x050 (R) - Memory Module Feature Register
	__I uint32_t ISAR[5]; // Offset: 0x060 (R) - Instruction Set Attributes Register
	uint32_t RESERVED0[5];
	__IO uint32_t CPACR; // Offset: 0x088 (R/W) - Coprocessor Access Control Register
} SCB_t;

/* SCB Vector Table Offset Register Definitions */
#define SCB_VTOR_TBLOFF_Pos 7
#define SCB_VTOR_TBLOFF_Msk (0x1FFFFFFUL << SCB_VTOR_TBLOFF_Pos)

/* Structure type to access the Core Debug Registers (CoreDebug). */
typedef struct {
	__IO uint32_t DHCSR; // Offset: 0x00 (R/W) - Debug Halting Control and Status Register
	__O uint32_t DCRSR; // Offset: 0x04 (W) - Debug Core Register Selector Register
	__IO uint32_t DCRDR; // Offset: 0x08 (R/W) - Debug Core Register Data Register
	__IO uint32_t DEMCR; // Offset: 0x0C (R/W) - Debug Exception and Monitor Control Register
} CoreDebug_t;

/* Debug Exception and Monitor Control Register Definitions */
#define CoreDebug_DEMCR_TRCENA_Pos 24
#define CoreDebug_DEMCR_TRCENA_Msk (1UL << CoreDebug_DEMCR_TRCENA_Pos)

/* Memory Mapping of Cortex-M4 Hardware */
#define SCS_Base (0xE000E000UL) // System Control Space Base Address
#define CoreDebug_Base (0xE000EDF0UL) // Core Debug Base Address
#define NVIC_Base (SCS_Base + 0x0100UL) // NVIC Base Address
#define SCB_Base (SCS_Base + 0x0D00UL) // System Control Block Base Address
#define SCB ((SCB_t*) SCB_Base) // SCB Configuration Structure
#define NVIC ((NVIC_t*) NVIC_Base) // NVIC Configuration Structure
#define CoreDebug ((CoreDebug_t*) CoreDebug_Base) // Core Debug Configuration Structure

/** HARDWARE ABSTRACTION LAYER **/

static inline void NVIC_DisableIRQ(IRQn_t irq) {
	NVIC->ICER[(uint32_t)((int32_t) irq) >> 5] = (uint32_t) (1 << ((uint32_t)((int32_t) irq) & (uint32_t) 0x1F));
}

#endif
