#include "nrf52.h"
#include "system.h"

static uint8_t errata12(void);
static uint8_t errata16(void);
static uint8_t errata31(void);
static uint8_t errata32(void);
static uint8_t errata36(void);
static uint8_t errata37(void);
static uint8_t errata57(void);
static uint8_t errata66(void);
static uint8_t errata108(void);
static uint8_t errata136(void);

uint32_t SystemCoreClock __attribute__((used)) = SYSTEM_CLOCK_64M;

void SystemInit(void) {
	// workaround for errata 12: "Comparator: Reference ladder not correctly calibrated."
	if (errata12()) {
		*(volatile uint32_t*) 0x40013540 = (*(uint32_t*) 0x10000324 & 0x00001F00) >> 8;
	}

	// workaround for errata 16: "System: RAM may be corrupt on wakeup from CPU IDLE."
	if (errata16()) {
		*(volatile uint32_t*) 0x4007C074 = 3131961357UL;
	}

	// workaround for errata 31: "Clock: Calib. values not correctly loaded from FICR at reset."
	if (errata31()) {
		*(volatile uint32_t*) 0x4000053C = ((*(volatile uint32_t*) 0x10000244) & 0x0000E000) >> 13;
	}

	// workaround for errata 32: "Debug IF: Debug session automatically enable TracePort pins."
	if (errata32()) {
		CoreDebug->DEMCR &= ~CoreDebug_DEMCR_TRCENA_Msk;
	}

	// workaround for errata 36: "Clock: Some registers are not reset when expected."
	if (errata36()) {
		CLOCK->EVENTS_DONE = 0;
		CLOCK->EVENTS_CTTO = 0;
		CLOCK->CTIV = 0;
	}

	// workaround for errata 37: "Radio: Encryption engine is slow by default."
	if (errata37()) {
		*(volatile uint32_t*) 0x400005A0 = 0x03;
	}

	// workaround for errata 57: "NFC: NFC modulation amplitude."
	if (errata57()) {
		*(volatile uint32_t*) 0x40005610 = 0x00000005;
		*(volatile uint32_t*) 0x40005688 = 0x00000001;
		*(volatile uint32_t*) 0x40005618 = 0x00000000;
		*(volatile uint32_t*) 0x40005614 = 0x0000003F;
	}

	// workaround for errata 66: "Temp: Linearity specification not met with default settings."
	if (errata66()) {
		TEMP->A0 = FICR->TEMPERATURE.A0;
		TEMP->A1 = FICR->TEMPERATURE.A1;
		TEMP->A2 = FICR->TEMPERATURE.A2;
		TEMP->A3 = FICR->TEMPERATURE.A3;
		TEMP->A4 = FICR->TEMPERATURE.A4;
		TEMP->A5 = FICR->TEMPERATURE.A5;
		TEMP->B0 = FICR->TEMPERATURE.B0;
		TEMP->B1 = FICR->TEMPERATURE.B1;
		TEMP->B2 = FICR->TEMPERATURE.B2;
		TEMP->B3 = FICR->TEMPERATURE.B3;
		TEMP->B4 = FICR->TEMPERATURE.B4;
		TEMP->B5 = FICR->TEMPERATURE.B5;
		TEMP->T0 = FICR->TEMPERATURE.T0;
		TEMP->T1 = FICR->TEMPERATURE.T1;
		TEMP->T2 = FICR->TEMPERATURE.T2;
		TEMP->T3 = FICR->TEMPERATURE.T3;
		TEMP->T4 = FICR->TEMPERATURE.T4;
	}

	// workaround for errata 108: "RAM: RAM content can't be trusted on wake from ON Idle or OFF."
	if (errata108()) {
		*(volatile uint32_t*) 0x40000EE4 = *(volatile uint32_t*) 0x10000258 & 0x0000004F;
	}

	// workaround for errata 136: "System: Bits in RESETREAS are set when they should not be."
	if (errata136()) {
		if (POWER->RESETREAS & POWER_RESETREAS_RESETPIN_Msk) {
			POWER->RESETREAS = ~POWER_RESETREAS_RESETPIN_Msk;
		}
	}

	// enable the FPU
	SCB->CPACR |= (3UL << 20) | (3UL << 22);
	__DSB();
	__ISB();

	// update the clock
	SystemCoreClock = SYSTEM_CLOCK_64M;
}

static uint8_t errata12(void) {
	if ((((*(uint32_t*) 0xF0000FE0) & 0x000000FF) == 0x06) && (((*(uint32_t*) 0xF0000FE4) & 0x0000000F) == 0x00)) {
		if (((*(uint32_t*) 0xF0000FE8) & 0x000000F0) == 0x30) {
			return 1;
		}

		if (((*(uint32_t*) 0xF0000FE8) & 0x000000F0) == 0x40) {
			return 1;
		}

		if (((*(uint32_t*) 0xF0000FE8) & 0x000000F0) == 0x50) {
			return 1;
		}
	}

	return 0;
}

static uint8_t errata16(void) {
	if ((((*(uint32_t*) 0xF0000FE0) & 0x000000FF) == 0x06) && (((*(uint32_t*) 0xF0000FE4) & 0x0000000F) == 0x00)) {
		if (((*(uint32_t*) 0xF0000FE8) & 0x000000F0) == 0x30) {
			return 1;
		}
	}

	return 0;
}

static uint8_t errata31(void) {
	if ((((*(uint32_t*) 0xF0000FE0) & 0x000000FF) == 0x06) && (((*(uint32_t*) 0xF0000FE4) & 0x0000000F) == 0x00)) {
		if (((*(uint32_t*) 0xF0000FE8) & 0x000000F0) == 0x30) {
			return 1;
		}

		if (((*(uint32_t*) 0xF0000FE8) & 0x000000F0) == 0x40) {
			return 1;
		}

		if (((*(uint32_t*) 0xF0000FE8) & 0x000000F0) == 0x50) {
			return 1;
		}
	}

	return 0;
}

static uint8_t errata32(void) {
	if ((((*(uint32_t*) 0xF0000FE0) & 0x000000FF) == 0x06) && (((*(uint32_t*) 0xF0000FE4) & 0x0000000F) == 0x00)) {
		if (((*(uint32_t*) 0xF0000FE8) & 0x000000F0) == 0x30) {
			return 1;
		}
	}

	return 0;
}

static uint8_t errata36(void) {
	if ((((*(uint32_t*) 0xF0000FE0) & 0x000000FF) == 0x06) && (((*(uint32_t*) 0xF0000FE4) & 0x0000000F) == 0x00)) {
		if (((*(uint32_t*) 0xF0000FE8) & 0x000000F0) == 0x30) {
			return 1;
		}

		if (((*(uint32_t*) 0xF0000FE8) & 0x000000F0) == 0x40) {
			return 1;
		}

		if (((*(uint32_t*) 0xF0000FE8) & 0x000000F0) == 0x50) {
			return 1;
		}
	}

	return 0;
}

static uint8_t errata37(void) {
	if ((((*(uint32_t*) 0xF0000FE0) & 0x000000FF) == 0x06) && (((*(uint32_t*) 0xF0000FE4) & 0x0000000F) == 0x00)) {
		if (((*(uint32_t*) 0xF0000FE8) & 0x000000F0) == 0x30) {
			return 1;
		}
	}

	return 0;
}

static uint8_t errata57(void) {
	if ((((*(uint32_t*) 0xF0000FE0) & 0x000000FF) == 0x06) && (((*(uint32_t*) 0xF0000FE4) & 0x0000000F) == 0x00)) {
		if (((*(uint32_t*) 0xF0000FE8) & 0x000000F0) == 0x30) {
			return 1;
		}
	}

	return 0;
}

static uint8_t errata66(void) {
	if ((((*(uint32_t*) 0xF0000FE0) & 0x000000FF) == 0x06) && (((*(uint32_t*) 0xF0000FE4) & 0x0000000F) == 0x00)) {
		if (((*(uint32_t*) 0xF0000FE8) & 0x000000F0) == 0x50) {
			return 1;
		}
	}

	return 0;
}

static uint8_t errata108(void) {
	if ((((*(uint32_t*) 0xF0000FE0) & 0x000000FF) == 0x06) && (((*(uint32_t*) 0xF0000FE4) & 0x0000000F) == 0x00)) {
		if (((*(uint32_t*) 0xF0000FE8) & 0x000000F0) == 0x30) {
			return 1;
		}

		if (((*(uint32_t*) 0xF0000FE8) & 0x000000F0) == 0x40) {
			return 1;
		}

		if (((*(uint32_t*) 0xF0000FE8) & 0x000000F0) == 0x50) {
			return 1;
		}
	}

	return 0;
}

static uint8_t errata136(void) {
	if ((((*(uint32_t*) 0xF0000FE0) & 0x000000FF) == 0x06) && (((*(uint32_t*) 0xF0000FE4) & 0x0000000F) == 0x00)) {
		if (((*(uint32_t*) 0xF0000FE8) & 0x000000F0) == 0x30) {
			return 1;
		}

		if (((*(uint32_t*) 0xF0000FE8) & 0x000000F0) == 0x40) {
			return 1;
		}

		if (((*(uint32_t*) 0xF0000FE8) & 0x000000F0) == 0x50) {
			return 1;
		}
	}

	return 0;
}
